import logging
import itertools

from hm_constants import CURRENT_TIME_DAY, CURRENT_TIME_HOUR, CURRENT_TIME_MIN

#mapping for chunks of heating schedule for a day
MAP_HOUR = 0
MAP_MIN = 1
MAP_TEMP = 2
#hour that indicates unused item
HOUR_UNUSED = 24
#index for returned schedule
SCH_ENT_DAY = 0
SCH_ENT_HOUR = 1
SCH_ENT_MIN = 2
SCH_ENT_TEMP = 3
#useful constants
HOUR_MINUTES = 60

class scheduler(object):
  #entry is a day or week/end, item is a part within the entry
  fieldbase = None
  fieldnames = ['mon','tues','wed','thurs','fri','sat','sun','wday','wend']
  
  def __init__(self):
    if not self.fieldbase is None:
      self.entrynames = [x + self.fieldbase for x in self.entrynames]
      self.fieldnames = [x + self.fieldbase for x in self.fieldnames]
    self.entries = dict.fromkeys(self.fieldnames, None)
    
  def set_raw_all(self, schedule):
    for entry in self.fieldnames:
      self.set_raw(entry, schedule)

  def set_raw(self, entry, schedule):
    if not entry in self.fieldnames:
      raise ValueError('Schedule entry does not exist %s'%entry)
    if not len(schedule) is self.valuesperentry * self.entriesperday:
      raise ValueError('Schedule entry wrong length %i'%len(entry))
    self.entries[entry] = schedule
    
  def pad_schedule(self,schedule):
    if not len(schedule)%self.valuesperentry == 0:
      raise IndexError("Schedule length not multiple of %d"%self.valuesperentry)
    pad_item = [HOUR_UNUSED,0,12][0:self.valuesperentry]
  
    return schedule + pad_item * ((self.valuesperentry * self.entriesperday - len(schedule))/self.valuesperentry)
    
  def display(self):
    print self.title + " Schedule"
    
    for name, entry in itertools.izip(self.printnames,self.entrynames):
      if self.entries[entry] is None:
        textstr = "None"
      else:
        textstr = self.entry_text(self.entries[entry])
      print(name.ljust(10) + textstr)
      logging.info(textstr)
      
  def _chunks(self, l, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i:i + n]
  
  def _reversechunks(self, l, n):
    """Yield successive n-sized chunks from l."""
    for i in range(len(l)-n, -1, -n):
        yield l[i:i + n]
      
  def getCurrentScheduleItem(self,timearray):
    ####check time and vars current
    
    todayschedule = self._getScheduleEntry(timearray[CURRENT_TIME_DAY])
      
    scheduletarget = self._getCurrentItemFromAnEntry(todayschedule,timearray)
      
    if scheduletarget == None:
      yestschedule = self._getPreviousScheduleEntry(timearray)
      scheduletarget = self._getLastItemFromAnEntry(yestschedule)
      return [self._getPreviousDay(timearray)] + scheduletarget
    else:
      return [timearray[CURRENT_TIME_DAY]] + scheduletarget
      
  def getNextScheduleItem(self,timearray):

    todayschedule = self._getScheduleEntry(timearray[CURRENT_TIME_DAY])
    
    scheduletarget = self._getNextItemFromAnEntry(todayschedule,timearray)
      
    if scheduletarget == None:
      tomschedule = self._getNextScheduleEntry(timearray)
      scheduletarget = self._getFirstItemFromAnEntry(tomschedule)
      return [self._getNextDay(timearray)] + scheduletarget
    else:
      return [timearray[CURRENT_TIME_DAY]] + scheduletarget
      
  def _getNextItemFromAnEntry(self,schedule,timearray):
    scheduletarget = None
    dayminutes = timearray[CURRENT_TIME_HOUR] * HOUR_MINUTES + timearray[CURRENT_TIME_MIN]

    for i in self._reversechunks(schedule,self.valuesperentry):
      if dayminutes < i[MAP_HOUR] * HOUR_MINUTES + i[MAP_MIN] and i[MAP_HOUR] != HOUR_UNUSED:
        scheduletarget = i
    
    return scheduletarget
  
  def _getCurrentItemFromAnEntry(self,schedule,timearray):
    scheduletarget = None
    dayminutes = timearray[CURRENT_TIME_HOUR] * HOUR_MINUTES + timearray[CURRENT_TIME_MIN]
    
    for i in self._chunks(schedule,self.valuesperentry):
      if dayminutes >= i[MAP_HOUR] * HOUR_MINUTES + i[MAP_MIN] and i[MAP_HOUR] != HOUR_UNUSED:
        scheduletarget = i
    
    return scheduletarget
    
  def _getPreviousScheduleEntry(self,timearray):
    return self._getScheduleEntry(self._getPreviousDay(timearray))

  def _getNextScheduleEntry(self,timearray):
    return self._getScheduleEntry(self._getNextDay(timearray))

  def _getPreviousDay(self,timearray):
    #shift from 1-7 to 0-6, subtract 1, modulo, shift back to 1-7
    return ((timearray[CURRENT_TIME_DAY] - 1 - 1 ) % 7 ) + 1
    
  def _getNextDay(self,timearray):
    #shift from 1-7 to 0-6, add 1, modulo, shift back to 1-7
    return ((timearray[CURRENT_TIME_DAY] - 1 + 1 ) % 7 ) + 1
    
  def _getFirstItemFromAnEntry(self,schedule):
    #gets first schedule entry if valid (not 24)
    firstentry = self._chunks(schedule,self.valuesperentry).next()
    if firstentry[MAP_HOUR] != HOUR_UNUSED:
      return firstentry
    else:
      return None

  def _getLastItemFromAnEntry(self,schedule):
    #gets last valid schedule entry (not 24)
    scheduletarget = None
    for i in self._reversechunks(schedule,self.valuesperentry):
          if i[MAP_HOUR] != HOUR_UNUSED:
            scheduletarget = i
            break
    
    return scheduletarget

class schedulerday(scheduler):
  entrynames = ['mon','tues','wed','thurs','fri','sat','sun']
  printnames = ['Monday','Tuesday','Wednesday','Thursday','Friday','Saturday','Sunday']
  
  def _getScheduleEntry(self, day):
    return self.entries[self.entrynames[day - 1]]

class schedulerweek(scheduler):
  entrynames = ['wday','wend']
  printnames = ['Weekdays','Weekends']
  
  def _getScheduleEntry(self, day):
    if day == 6 or day == 7:
      return self.entries[self.entrynames[1]]
    elif day >= 1 and day <= 5:
      return self.entries[self.entrynames[0]]
    else:
      raise ValueError("Day not recognised")
      
class schedulerheat(scheduler):
  title = 'Heating'
  valuesperentry = 3
  entriesperday = 4
  fieldbase = '_heat'
  
  def entry_text(self, data):
    tempstr = ''
    for set in self._chunks(data,self.valuesperentry):
      if set[MAP_HOUR] != HOUR_UNUSED:
        tempstr += "%02d:%02d at %02iC " % (set[MAP_HOUR],set[MAP_MIN],set[MAP_TEMP])
        
    return tempstr
  
class schedulerwater(scheduler):
  title = 'Hot Water'
  valuesperentry = 2
  entriesperday = 8
  fieldbase = '_water'
  
  def entry_text(self, data):
    toggle = True
    count = 1
  
    tempstr = ''
    for set in self._chunks(data,2):
      if set[0] != HOUR_UNUSED:
        if toggle:
          tempstr += "Time %i On at %02d:%02d " % (count,set[0],set[1])
        else:
          tempstr += "Off at %02d:%02d, " % (set[0],set[1])
          count=count+1
        toggle = not toggle
        
    return tempstr
  
class schedulerdayheat(schedulerday, schedulerheat):
  pass
class schedulerweekheat(schedulerweek, schedulerheat):
  pass
class schedulerdaywater(schedulerday, schedulerwater):
  pass
class schedulerweekwater(schedulerweek, schedulerwater):
  pass



#
# Ian Horsley 2018

#
# function to setup logging to screen and files

# Assume Python 2.7.x
#
import logging
import os

from logging.handlers import RotatingFileHandler

def initialize_logger(output_dir, screenlevel, debugLog = None):
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)
     
    # create console handler and set level to info
    handler = logging.StreamHandler()
    handler.setLevel(screenlevel)
    formatter = logging.Formatter("%(name)s - %(levelname)s - %(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    # create error file handler and set level to error
    handler = logging.FileHandler(os.path.join(output_dir, "error.log"),"w", encoding=None, delay="true")
    handler.setLevel(logging.WARN)
    formatter = logging.Formatter("%(asctime)-15s %(levelname)s - %(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    #a is append, w is write
    if debugLog != None:
      #create debug file handler and set level to debug
      handler = RotatingFileHandler(os.path.join(output_dir, "all.log"), mode='a', maxBytes=5*1024*1024, 
                                   backupCount=2, encoding=None, delay=0)
      #handler = logging.FileHandler(os.path.join(output_dir, "all.log"),"w", encoding=None, delay="true")
      handler.setLevel(logging.DEBUG)
      formatter = logging.Formatter("%(asctime)-15s %(levelname)s - %(message)s")
      handler.setFormatter(formatter)
      logger.addHandler(handler)


#
# Ian Horsley 2018

#
# hmNetwork Class and helper functions
# also loads adaptor and keeps track of heatmiser devices

# Assume Python 2.7.x
#
import os
import serial
import time
import logging
import sys

# Import our own stuff
from devices import *
from adaptor import *
import setup as hms

class HeatmiserNetwork:
### stat list setup

  def __init__(self, configfile = None):
    
    # Select default configuration file if none provided
    if configfile is None:
      self._module_path = os.path.abspath(os.path.dirname(__file__))
      configfile = os.path.join(self._module_path, "hmcontroller.conf")
      
    # Initialize controller setup
    try:
        self._setup = hms.HeatmiserControllerFileSetup(configfile)
        settings = self._setup.settings
    except hms.HeatmiserControllerSetupInitError as e:
        logging.error(e)
        sys.exit("Unable to load configuration file: " + configfile)
    
    # Initialize and connect to heatmiser network, probably through serial port
    self.adaptor = Heatmiser_Adaptor(self._setup)
    self.adaptor.connect()
    
    # Create a broadcast device
    setattr(self,"All",hmBroadcastController(self.adaptor,"Broadcast to All"))
    self._current = self.All
    
    self.setStatList(settings['devices'], settings['devicesgeneral'])
      
  def setStatList(self, list, generalsettings):
    # Store list of stats
    self._statlist = list
    self._statnum = len(self._statlist)

    self.controllers = range(self._statnum)
    for name, controllersettings in list.iteritems():
      controllersettings['name'] = name
      if hasattr(self,name):
        print "error duplicate stat name"
      else:
        setattr(self,name,hmController(self.adaptor,controllersettings, generalsettings))
        self.controllers[controllersettings['display_order']-1] = getattr(self,name)

    self._current = self.controllers[0]
  
  def getStatAddress(self,shortname):
    if isinstance(shortname,basestring):
      #matches = [x for x in self.statlist if x[SL_SHORT_NAME] == shortname]
      shorts = [row[SL_SHORT_NAME] for row in self._statlist]
      
      return self.statlist[shorts.index(shortname)]['address']
    else:
      return shortname

  def setCurrentControllerByName(self,name):
    self._current = getattr(self,name)
    
  def setCurrentControllerByIndex(self,index):
    self._current = self.controllers[index]
    
  def controllerByName(self,name):
    return getattr(self,name)
    
  def on_all(self, method, *args, **kwargs):
    results = []
    for obj in self.controllers:
        results.append(getattr(obj, method)(*args, **kwargs))
    return results



#!/usr/bin/env python
#
# Ian Horsley 2018

#
# Gets data for all stats repeatedly and displays
#
import time
import logging

from heatmisercontroller.logging_setup import initialize_logger
from heatmisercontroller.hm_constants import *
from heatmisercontroller.network import *
from heatmisercontroller.exceptions import hmResponseError

#start logging
initialize_logger('logs', logging.DEBUG, True)

hmn1 = HeatmiserNetwork()

# CYCLE THROUGH ALL CONTROLLERS
for current_controller in hmn1.controllers:
  print "\r\nGetting all data control %2d in %s *****************************" % (current_controller._address, current_controller._long_name)
  
  try:
    current_controller.hmReadAll()
  except hmResponseError as e:
    print "C%d in %s Failed to Read due to %s" % (current_controller._address,  current_controller._name.ljust(4), str(e))
  else:
    disptext = "C%d Air Temp is %.1f from type %.f and Target set to %d  Boiler Demand %d" % (current_controller._address, current_controller.getAirTemp(), current_controller.getAirSensorType(), current_controller.setroomtemp, current_controller.heatingdemand)
    if current_controller.isHotWater():
      print "%s Hot Water Demand %d" % (disptext, current_controller.hotwaterdemand)
    else:
      print disptext
    current_controller.display_heating_schedule()
    current_controller.display_water_schedule()

time.sleep(5) # sleep before next cycle

while True:
  print
  # CYCLE THROUGH ALL CONTROLLERS repeatedly
  for current_controller in hmn1.controllers:
    
    try:
      current_controller.hmReadTempsandDemand()
    except hmResponseError as e:
      print "C%d in %s Failed to Read due to %s" % (current_controller._address,  current_controller._name.ljust(4),str(e))
    else: 
      targettext = current_controller.printTarget()
      disptext = "C%d in %s Air Temp is %.1f from type %.f, %s, Heat %d" % (current_controller._address,  current_controller._name.ljust(4), current_controller.getAirTemp(), current_controller.getAirSensorType(), targettext, current_controller.heatingdemand)
      
      if current_controller.isHotWater():
        print "%s Water %d" % (disptext, current_controller.hotwaterdemand)
      else:
        print disptext

  time.sleep(5) # sleep before next cycle



#
# Ian Horsley 2018

#
# Heatmiser_Adaptor Class and helper functions
# handles serial connection and basic framing for the heatmiser protocol

# Assume Python 2.7.x
#

import time
import logging
import serial

from hm_constants import *
import framing
from .exceptions import hmResponseError, hmResponseErrorCRC
 
def retryer(max_retries=3):
    def wraps(func):

        def inner(*args, **kwargs):
            for i in range(max_retries):
                if i is not 0:
                    logging.warn("Gen retrying due to %s"%str(lasterror))
                try:        
                    result = func(*args, **kwargs)
                except hmResponseError as e:
                    lasterror = e
                    continue
                else:
                    return result
            else:
                raise hmResponseError("Failed after %i retries on %s"%(max_retries,str(lasterror))) 
        return inner
    return wraps
        
class Heatmiser_Adaptor(object):

    def __init__(self, setup):
    
        # Initialize setup and get settings
        self._setup = setup
        settings = self._setup.settings
        
        self.serport = serial.Serial()
        self.serport.bytesize = COM_SIZE = serial.EIGHTBITS
        self.serport.parity = COM_PARITY = serial.PARITY_NONE
        self.serport.stopbits = COM_STOP = serial.STOPBITS_ONE
        
        self.lastsendtime = None
        self.creationtime = time.time()
        
        self._update_settings(settings)
        
        self.lastreceivetime = self.creationtime - self.serport.COM_BUS_RESET_TIME # so that system will get on with sending straight away
        
    def __del__(self):
        self._disconnect()
        
###

    def _update_settings(self, settings):
        """Check settings and update if needed."""     
        
        for name, value in settings['controller'].iteritems():
            setattr(self, name, value)
        
        # Configure serial settings         
                
        wasopen = False
        if self.serport.isOpen():
            wasopen = True
            self.serport.close() # close port
        
        for name, value in settings['serial'].iteritems():
            setattr(self.serport, name, value)
        
        if not self.serport.isOpen() and wasopen:
            try:
                self.serport.open()
            except serial.SerialException as e:
                logging.error("Could not open serial port %s: %s" % (self.serport.portstr, e))
                raise
                
### low level serial commands

    def connect(self):
        if not self.serport.isOpen():
            try:
                self.serport.open()
            except serial.SerialException as e:
                logging.error("Could not open serial port %s: %s" % (self.serport.portstr, e))
                raise

            logging.info("Gen %s port opened"% (self.serport.portstr))
            logging.debug("Gen %s baud, %s bit, %s parity, with %s stopbits, timeout %s seconds" % (self.serport.baudrate, self.serport.bytesize, self.serport.parity, self.serport.stopbits, self.serport.timeout))
        else:
            logging.warn("Gen serial port was already open")
        
    def _disconnect(self):
        #check if serial port is open and if so close
        #shouldn't need to call because handled by destructor
        if self.serport.isOpen():
            self.serport.close() # close port
            logging.info("Gen serial port closed")
        else:
            logging.warn("Gen serial port was already closed")
            
    def _hmSendMsg(self, message) :
            if not self.serport.isOpen():
                self.connect()

            #check time since last received to make sure bus has settled.
            waittime = self.serport.COM_BUS_RESET_TIME - (time.time() - self.lastreceivetime)
            if waittime > 0:
                logging.debug("Gen waiting before sending %.2f"% ( waittime ))
                time.sleep(waittime)
            
            # http://stackoverflow.com/questions/180606/how-do-i-convert-a-list-of-ascii-values-to-a-string-in-python
            string = ''.join(map(chr,message))

            try:
                written = self.serport.write(string)    # Write a string
            except serial.SerialTimeoutException as e:
                self.serport.close() #need to close so that isOpen works correctly.
                logging.warning("Write timeout error: %s, sending %s" % (e, ', '.join(str(x) for x in message)))
                raise
            except serial.SerialException as e:
                self.serport.close() #need to close so that isOpen works correctly.
                logging.warning("Write error: %s, sending %s" % (e,    ', '.join(str(x) for x in message)))
                raise
            else:
                self.lastsendtime = time.strftime("%d %b %Y %H:%M:%S +0000", time.localtime(time.time())) #timezone is wrong
                logging.debug("Gen sent %s",', '.join(str(x) for x in message))

    def _hmClearInputBuffer(self):
        #clears input buffer
        #use after CRC check wrong; encase more data was sent than expected.
    
        time.sleep(self.serport.COM_TIMEOUT) #wait for read timeout to ensure slave finished sending
        try:
            if self.serport.isOpen():
                self.serport.reset_input_buffer() #reset input buffer and dump any contents
            logging.warning("Input buffer cleared")
        except serial.SerialException as e:
            self.serport.close()
            logging.warning("Failed to clear input buffer")
            raise
                    
    def _hmReceiveMsg(self, length = MAX_FRAME_RESP_LENGTH) :
            # Listen for a reply
            if not self.serport.isOpen():
                self.connect()
            logging.debug("Gen listening for %d"%length)
            
            # Listen for the first byte
            timereadstart = time.time()
            self.serport.timeout = self.serport.COM_START_TIMEOUT #wait for start of response
            try:
                firstbyteread = self.serport.read(1)
            except serial.SerialException as e:
                #There is no new data from serial port (or port missing) (Doesn't include no response from stat)
                logging.warning("Gen serial port error: %s" % str(e))
                self.serport.close()
                raise
            else:
                timereadfirstbyte = time.time()-timereadstart
                logging.debug("Gen waited %.2fs for first byte"%timereadfirstbyte)
                if len(firstbyteread) == 0:
                    raise hmResponseError("No Response")
                
                # Listen for the rest of the response
                self.serport.timeout = max(self.serport.COM_MIN_TIMEOUT, self.serport.COM_TIMEOUT - timereadfirstbyte) #wait for full time out for rest of response, but not less than COM_MIN_TIMEOUT)
                try:
                    byteread = self.serport.read(length - 1)
                except serial.SerialException as e:
                    #There is no new data from serial port (or port missing) (Doesn't include no response from stat)
                    logging.warning("Gen serial port error: %s" % str(e))
                    self.serport.close()
                    raise

                #Convert back to array
                data = map(ord,firstbyteread) + map(ord,byteread)

                return data
            finally:
                self.serport.timeout = self.serport.COM_TIMEOUT #make sure timeout is reverted
                self.lastreceivetime = time.time() #record last read time. Used to manage bus settling.

### protocol functions
    
    @retryer(max_retries = 3)
    def hmWriteToController(self, network_address, protocol, dcb_address, length, payload):
        ###shouldn't be labelled dcb_address. It is a unique address.
        msg = framing._hmFormFrame(network_address, protocol, self.my_master_addr, FUNC_WRITE, dcb_address, length, payload)
        
        try:
            self._hmSendMsg(msg)
        except Exception as e:
            logging.warn("C%i writing to address, no message sent"%(network_address))
            raise
        else:
            logging.debug("C%i written to address %i length %i payload %s"%(network_address,dcb_address, length, ', '.join(str(x) for x in payload)))
            if network_address == BROADCAST_ADDR:
                self.lastreceivetime = time.time() + self.serport.COM_SEND_MIN_TIME - self.serport.COM_BUS_RESET_TIME # if broadcasting force it to wait longer until next send
            else:
                response = self._hmReceiveMsg(FRAME_WRITE_RESP_LENGTH)
                try:
                    framing._hmVerifyWriteAck(protocol, network_address, self.my_master_addr, response)
                except hmResponseErrorCRC:
                    self._hmClearInputBuffer()
                    raise
                    
    def minTimeBetweenReads(self):
        return self.serport.COM_BUS_RESET_TIME

    @retryer(max_retries = 2)
    def hmReadFromController(self, network_address, protocol, dcb_start_address, expectedLength, readall = False):
        ###mis labelled dcb addres, should be unique
            if readall:
                msg = framing._hmFormReadFrame(network_address, protocol, self.my_master_addr, DCB_START, RW_LENGTH_ALL)
                logging.debug("C %i read request to address %i length %i"%(network_address,DCB_START, RW_LENGTH_ALL))
            else:
                msg = framing._hmFormReadFrame(network_address, protocol, self.my_master_addr, dcb_start_address, expectedLength)
                logging.debug("C %i read request to address %i length %i"%(network_address,dcb_start_address, expectedLength))
            
            try:
                self._hmSendMsg(msg)
            except:
                logging.warn("C%i address, read message not sent"%(network_address))
                raise
            else:
                time1 = time.time()

                try:
                    response = self._hmReceiveMsg(MIN_FRAME_READ_RESP_LENGTH + expectedLength)
                except Exception as e:
                    logging.warn("C%i read failed from address %i length %i due to %s"%(network_address,dcb_start_address, expectedLength, str(e)))
                    raise
                else:
                    logging.debug("C%i read in %.2f s from address %i length %i response %s"%(network_address,time.time()-time1,dcb_start_address, expectedLength, ', '.join(str(x) for x in response)))
                
                    try:
                        framing._hmVerifyResponse(protocol, network_address, self.my_master_addr, FUNC_READ, expectedLength , response)
                    except hmResponseErrorCRC:
                        self._hmClearInputBuffer()
                        raise
                    return response[FR_CONTENTS:-CRC_LENGTH]

    def hmReadAllFromController(self, network_address, protocol, expectedLength):
        return self.hmReadFromController(network_address, protocol, DCB_START, expectedLength, True)
        




"""


"""

import time
import logging
import os
from configobj import ConfigObj
from validate import Validator

from .exceptions import HeatmiserControllerSetupInitError

"""class HeatmiserControllerSetup

User interface to setup the contoller.

The settings attribute stores the settings of the hub. It is a
dictionary with the following keys:

        'controller': a dictionary containing general settings
        'serial': a dictionary containing the serial port settings
        'devices': a dictionary containing the configuration of each remote device
        'setup': 

        The controller settings are:
        'loglevel': the logging level
        
        ###interfacers and reporters are dictionaries with the following keys:
        ###'Type': class name
        ###'init_settings': dictionary with initialization settings
        ###'runtimesettings': dictionary with runtime settings
        ###Initialization and runtime settings depend on the interfacer and
        ###reporter type.

The run() method is supposed to be run regularly by the instantiator, to
perform regular communication tasks.

The check_settings() method is run regularly as well. It checks the settings 
and returns True is settings were changed.

This almost empty class is meant to be inherited by subclasses specific to
each setup.

"""

class HeatmiserControllerSetup(object):

    def __init__(self):
        
        # Initialize logger
        self._log = logging.getLogger("HeatmiserController")
        
        # Initialize settings
        self.settings = None

    def run(self):
        """Run in background. 
        
        To be implemented in child class.

        """
        pass

    def check_settings(self):
        """Check settings
        
        Update attribute settings and return True if modified.
        
        To be implemented in child class.
        
        """

class HeatmiserControllerFileSetup(HeatmiserControllerSetup):

    def __init__(self, filename):
        
        # Initialization
        super(HeatmiserControllerFileSetup, self).__init__()

        self._module_path = os.path.abspath(os.path.dirname(__file__))
        specpath = os.path.join(self._module_path, "hmcontroller.spec")
        
        # List of expected sections
        self._sections = ['controller', 'serial', 'devices', 'setup']
            
        # Initialize attribute settings as a ConfigObj instance
        logging.debug("Loading %s and checking against %s"%(filename, specpath))
        try:
            self.settings = ConfigObj(filename, file_error=True, configspec=specpath)
            self._validator = Validator()
        except IOError as e:
            raise HeatmiserControllerSetupInitError(e)
        except SyntaxError as e:
            raise HeatmiserControllerSetupInitError(
                'Error parsing config file \"%s\": ' % filename + str(e))
        except KeyError as e:
            raise HeatmiserControllerSetupInitError(
                'Configuration file error - section missing: ' + str(e))
        
        #check settings and add any default values
        self._check_settings()
        
        # Initialize update timestamps
        self._settings_update_timestamp = 0
        
        # Load use configured variables
        for name, value in self.settings['setup'].iteritems():
            setattr(self, '_c_'+name, value)

        # create a timeout message if time out is set (>0)
        self.retry_msg = " Retry in " + str(self._c_retry_time_interval) + " seconds" if self._c_retry_time_interval <= 0 else ""

    def reload_settings(self):
        """Reload and check settings

        Update attribute settings and return True if modified. Return False if failed to load new settings. Return None if nothing new or not checked

        """
        
        # Check settings only once per second
        now = time.time()
        if now - self._settings_update_timestamp < 0:
            return
        # Update timestamp
        self._settings_update_timestamp = now + self._c_check_time_interval
        
        # Backup settings
        settings = dict(self.settings)
        
        # Get settings from file
        try:
            self.settings.reload()
        except IOError as e:
            self._log.warning('Could not get settings: ' + str(e) + self.retry_msg)
            self._settings_update_timestamp = now + self._c_retry_time_interval
            return
        except SyntaxError as e:
            self._log.warning('Could not get settings: ' + 
                              'Error parsing config file: ' + str(e) + self.retry_msg)
            self._settings_update_timestamp = now + self._c_retry_time_interval
            return
        except Exception:
            import traceback
            self._log.warning("Couldn't get settings, Exception: " +
                              traceback.format_exc() + self.retry_msg)
            self._settings_update_timestamp = now + self._c_retry_time_interval
            return

        if self.settings != settings:
            try:
                self._check_settings()
            except (ValueError, KeyError):
                self.settings = settings
                return False
            return True
            
    def _check_settings(self):
        try:
            # Validate the configuration file and copy any missing defaults
            returnval = self.settings.validate(self._validator, preserve_errors=True, copy=True)
            if not returnval is True:
                raise ValueError("failed validation of config file %s"%str(returnval))

            # Check the settings file sections exist
            for name in self._sections:
                self.settings[name]
        except (ValueError, KeyError) as e:
            logging.warning("Configuration parse failed : " + str(e))
            raise


            






#
# Ian Horsley 2018

#
# hmController class
# handles all the DCB and data objects for each controller on the heatmiser network

# Assume Python 2.7.x
#

#relook at the setfield and setfields. Should they be in adaptor. Payload length checking should happen on all results.
#could clean up class setup. basic controller could be write only or read only. two classes that inherite. General, gets both, and broadcast, takes write only.

import logging
import time
import serial
from datetime import datetime

from hm_constants import *
from .exceptions import hmResponseError, hmControllerTimeError
from schedule_functions import schedulerdayheat, schedulerweekheat, schedulerdaywater, schedulerweekwater, SCH_ENT_TEMP

class hmController(object):
  ##Variables used by code
  lastreadtime = 0 #records last time of a successful read

  def __init__(self, adaptor, devicesettings, generalsettings = None):
    #address, protocol, short_name, long_name, model, mode
    self._adaptor = adaptor
    
    self.water_schedule = None
        
    #initialise data structures
    self._buildfieldtables()
    self.data = dict.fromkeys(self._fieldnametonum.keys(),None)
    self.datareadtime = dict.fromkeys(self._fieldnametonum.keys(),None)
    
    self._update_settings(devicesettings, generalsettings)

    self.rawdata = [None] * self.DCBlength
  
  def _update_settings(self, settings, generalsettings):
    """Check settings and update if needed."""   
    
    if not generalsettings is None:
        for name, value in generalsettings.iteritems():
            setattr(self, '_' + name, value)
    
    for name, value in settings.iteritems():
      setattr(self, '_' + name, value)
      
    if self._expected_prog_mode == PROG_MODE_DAY:
      self.heat_schedule = schedulerdayheat()
      if self.isHotWater():
        self.water_schedule = schedulerdaywater()
    elif self._expected_prog_mode == PROG_MODE_WEEK:
      self.heat_schedule = schedulerweekheat()
      if self.isHotWater():
        self.water_schedule = schedulerweekwater()
    else:
      raise ValueError("Unknown program mode")
    
    self._expected_prog_mode_number = PROG_MODES[self._expected_prog_mode]
    
    self._fieldranges = FIELDRANGES[self._expected_model][self._expected_prog_mode]
    
    if self._expected_model == 'prt_e_model':
      self.DCBmap = PRTEmap[self._expected_prog_mode]
    elif self._expected_model == 'prt_hw_model':
      self.DCBmap = PRTHWmap[self._expected_prog_mode]
    elif self._expected_model == False:
      self.DCBmap = STRAIGHTmap
    else:
      raise ValueError("Unknown model %s"%self._expected_model)

    self._buildDCBtables()
    
    self._expected_model_number = DEVICE_MODELS[self._expected_model]
    
    if self.DCBmap[0][1] != DCB_INVALID:
      self.DCBlength = self.DCBmap[0][0] - self.DCBmap[0][1] + 1
    elif self.DCBmap[1][1] != DCB_INVALID:
      self.DCBlength = self.DCBmap[1][0] - self.DCBmap[1][1] + 1
    else:
      raise ValueError("DCB map length not found")

    self.fullreadtime = self._estimateReadTime(self.DCBlength)
    
  def _getDCBaddress(self, uniqueaddress):
    #get the DCB address for a controller from the unique address
        return self._uniquetodcb[uniqueaddress]
  
  def _buildfieldtables(self):
    self._fieldnametonum = {}
    for key, data in enumerate(fields):
        fieldname = data[FIELD_NAME]
        self._fieldnametonum[fieldname] = key
        
  def _buildDCBtables(self):
    #build a forward lookup table for the DCB values from uniqueaddress
    self._uniquetodcb = range(MAX_UNIQUE_ADDRESS+1)
    for uniquemax, offsetsel in self.DCBmap:
        self._uniquetodcb[0:uniquemax + 1] = [x - offsetsel for x in range(uniquemax + 1)] if not offsetsel is DCB_INVALID else [DCB_INVALID] * (uniquemax + 1)
    
    #build list of valid fields for this stat
    self._fieldsvalid = [False] * len(fields)
    for first, last in self._fieldranges:
      self._fieldsvalid[self._fieldnametonum[first]: self._fieldnametonum[last] + 1] = [True] * (self._fieldnametonum[last] - self._fieldnametonum[first] + 1)
    #self._fullDCB = sum(x is not None for x in self._uniquetodcb))
    logging.debug("C%i Fieldsvalid %s"%(self._address,','.join(str(int(x)) for x in self._fieldsvalid)))
    
  def getRawData(self, startfieldname = None, endfieldname = None):
    if startfieldname == None or endfieldname == None:
      return self.rawdata
    else:
      return self.rawdata[self._getDCBaddress(uniadd[startfieldname][UNIADD_ADD]):self._getDCBaddress(uniadd[endfieldname][UNIADD_ADD])]
    
  def hmReadVariables(self):
    self.readFields('setroomtemp', 'hotwaterdemand')
    
  def hmReadTempsandDemand(self):
    self.readFields('remoteairtemp', 'hotwaterdemand')
    
  def readTime(self, maxage = 0):
    return self.readField('currenttime', maxage)

  def hmReadAll(self):
    try:
      self.rawdata = self._adaptor.hmReadAllFromController(self._address, self._protocol, self.DCBlength)
    except serial.SerialException as e:

      logging.warn("C%i Read all failed, Serial Port error %s"%(self._address, str(e)))
      raise
    else:
      logging.info("C%i Read all"%(self._address))

      self.lastreadtime = time.time()
      self._procpayload(self.rawdata)
      return self.rawdata

  def readField(self, fieldname, maxage = 0):
    #return field value
    #read field from network if
    # no maxage in request (maxage = 0)
    # maxage is valid and data too old
    # or not be read before (maxage = None)
    if maxage == 0 or (maxage is not None and not self._check_data_age(maxage, fieldname)) or not self._check_data_present(fieldname):
      if self._autoreadall is True:
        self.readFields(fieldname)
      else:
        raise ValueError("Need to read %s first"%fieldname)
    return self.data[fieldname]
    
  def _getFieldBlocks(self, firstfieldname, lastfieldname):
    #data can only be requested from the controller in contiguous blocks
    #functions takes a first and last field and seperates out the individual blocks avaliable for the controller type
    ###return, uniquestart, uniqueend, length of read
    #return, fieldstart, fieldend, length of read in bytes
    
    firstfieldid = self._fieldnametonum[firstfieldname]
    lastfieldid = self._fieldnametonum[lastfieldname]
    
    blocks = []
    previousfieldvalid = False

    for fieldnum, fieldvalid in enumerate(self._fieldsvalid[firstfieldid:lastfieldid + 1],firstfieldid):
        if previousfieldvalid is False and not fieldvalid is False:
            #start = fields[fieldnum][FIELD_ADD]
            start = fieldnum
        elif not previousfieldvalid is False and fieldvalid is False:
            #blocks.append([start,fields[fieldnum][FIELD_ADD],fields[fieldnum][FIELD_ADD] + fields[fieldnum][FIELD_LEN] - start])
            blocks.append([start,fieldnum - 1,fields[fieldnum - 1][FIELD_ADD] + fields[fieldnum - 1][FIELD_LEN] - fields[start][FIELD_ADD]])
        
        previousfieldvalid = fieldvalid

    if not previousfieldvalid is False:
        #blocks.append([start,fields[lastfieldid][FIELD_ADD],fields[lastfieldid][FIELD_ADD] + fields[lastfieldid][FIELD_LEN] - start])
        blocks.append([start,lastfieldid,fields[lastfieldid][FIELD_ADD] + fields[lastfieldid][FIELD_LEN] - fields[start][FIELD_ADD]])
    return blocks
  
  def _estimateBlocksReadTime(self,blocks):
    #estimates read time for a set of blocks, including the COM_BUS_RESET_TIME between blocks 
    #excludes the COM_BUS_RESET_TIME before first block
    readtimes = [self._estimateReadTime(x[2]) for x in blocks]
    return sum(readtimes) + self._adaptor.minTimeBetweenReads() * (len(blocks) - 1)
  
  def _estimateReadTime(self,length):
    #estiamtes the read time for a call to hmReadFromController without COM_BUS_RESET_TIME
    #based on empirical measurements of one prt_hw_model and 5 prt_e_model
    return length * 0.002075 + 0.070727
  
  def readFields(self, firstfieldname, lastfieldname = None):
    #reads fields from controller, safe for blocks crossing gaps in dcb
    if lastfieldname == None:
        lastfieldname = firstfieldname

    blockstoread = self._getFieldBlocks(firstfieldname, lastfieldname)
    logging.debug(blockstoread)
    estimatedreadtime = self._estimateBlocksReadTime(blockstoread)
    
    if estimatedreadtime < self.fullreadtime - 0.02: #if to close to full read time, then read all
        try:
            for firstfieldid, lastfieldid, blocklength in blockstoread:
                logging.debug("C%i Reading ui %i to %i len %i, proc %s to %s"%(self._address, fields[firstfieldid][FIELD_ADD],fields[lastfieldid][FIELD_ADD],blocklength,fields[firstfieldid][FIELD_NAME], fields[lastfieldid][FIELD_NAME]))
                rawdata = self._adaptor.hmReadFromController(self._address, self._protocol, fields[firstfieldid][FIELD_ADD], blocklength)
                self.lastreadtime = time.time()
                self._procpartpayload(rawdata, fields[firstfieldid][FIELD_NAME], fields[lastfieldid][FIELD_NAME])
        except serial.SerialException as e:
            logging.warn("C%i Read failed of fields %s to %s, Serial Port error %s"%(self._address, firstfieldname.ljust(FIELD_NAME_LENGTH),lastfieldname.ljust(FIELD_NAME_LENGTH), str(e)))
            raise
        else:
            logging.info("C%i Read fields %s to %s, in %i blocks"%(self._address, firstfieldname.ljust(FIELD_NAME_LENGTH),lastfieldname.ljust(FIELD_NAME_LENGTH),len(blockstoread)))
    else:
        logging.debug("C%i Read fields %s to %s by readAll, %0.3f %0.3f"%(self._address, firstfieldname.ljust(FIELD_NAME_LENGTH),lastfieldname.ljust(FIELD_NAME_LENGTH), estimatedreadtime, self.fullreadtime))
        self.hmReadAll()
  
  def _procfield(self,data,fieldinfo):
    fieldname = fieldinfo[FIELD_NAME]
    length = fieldinfo[FIELD_LEN]
    factor = fieldinfo[FIELD_DIV]
    range = fieldinfo[FIELD_RANGE]
    #logging.debug("Processing %s %s"%(fieldinfo[FIELD_NAME],', '.join(str(x) for x in data)))
    if length == 1:
      value = data[0]/factor
    elif length == 2:
      val_high = data[0]
      val_low  = data[1]
      value = 1.0*(val_high*256 + val_low)/factor #force float, although always returns integer temps.
    elif length == 4:
      value = data
    elif length == 12:
      self.heat_schedule.set_raw(fieldname,data)
      value = data
    elif length == 16:
      self.water_schedule.set_raw(fieldname,data)
      value = data
    else:
      raise ValueError("_procpayload can't process field length")
  
    if len(range) == 2 and isinstance(range[0], (int, long)) and isinstance(range[1], (int, long)):
      if value < range[0] or value > range[1]:
        raise hmResponseError("Field value %i outside expected range"%value)
    
    if fieldname == 'DCBlen' and value != self.DCBlength:
      raise hmResponseError('DCBlengh is unexpected')
    
    if fieldname == 'model' and value != self._expected_model_number:
      raise hmResponseError('Model is unexpected')
    
    if fieldname == 'programmode' and value != self._expected_prog_mode_number:
      raise hmResponseError('Programme mode is unexpected')
    
    if fieldname == 'version' and self._expected_model != 'prt_hw_model':
      value = data[0] & 0x7f
      self.floorlimiting = data[0] >> 7
      self.data['floorlimiting'] = self.floorlimiting
    
    self.data[fieldname] = value
    setattr(self, fieldname, value)
    self.datareadtime[fieldname] = self.lastreadtime
    
    if fieldname == 'currenttime':
      self._checkcontrollertime()
    
    ###todo, add range validation for other lengths

  def _procpartpayload(self, rawdata, firstfieldname, lastfieldname):
    #rawdata must be a list
    #converts field names to unique addresses to allow process of shortened raw data
    logging.debug("C%i Processing Payload from field %s to %s"%(self._address,firstfieldname,lastfieldname) )
    firstfieldid = self._fieldnametonum[firstfieldname]
    lastfieldid = self._fieldnametonum[lastfieldname]
    self._procpayload(rawdata, firstfieldid, lastfieldid)
    
  def _procpayload(self, rawdata, firstfieldid = 0, lastfieldid = len(fields)):
    logging.debug("C%i Processing Payload from field %i to %i"%(self._address,firstfieldid,lastfieldid) )

    fullfirstdcbadd = self._getDCBaddress(fields[firstfieldid][FIELD_ADD])
    
    for fieldinfo in fields[firstfieldid:lastfieldid + 1]:
      uniqueaddress = fieldinfo[FIELD_ADD]
      
      length = fieldinfo[FIELD_LEN]
      dcbadd = self._getDCBaddress(uniqueaddress)

      if dcbadd == DCB_INVALID:
        setattr(self, fieldinfo[FIELD_NAME], None)
        self.data[fieldinfo[FIELD_NAME]] = None
      else:
        dcbadd -= fullfirstdcbadd #adjust for the start of the request
        
        try:
          self._procfield(rawdata[dcbadd:dcbadd+length], fieldinfo)
        except hmResponseError as e:
          logging.warn("C%i Field %s process failed due to %s"%(self._address, fieldinfo[FIELD_NAME], str(e)))

    self.rawdata[fullfirstdcbadd:fullfirstdcbadd+len(rawdata)] = rawdata

  def _checkcontrollertime(self):
    #run compare of times, and try to fix if _autocorrectime
    try:
      self._comparecontrollertime()
    except hmControllerTimeError:
      if self._autocorrectime is True:
        self.setTime()
      else:
        raise
  
  def _comparecontrollertime(self):       
    # Now do same sanity checking
    # Check the time is within range
    # currentday is numbered 1-7 for M-S
    # localday (python) is numbered 0-6 for Sun-Sat
    
    if not self._check_data_present('currenttime'):
      raise hmResponseError("Time not read before check")

    localtimearray = self._localtimearray(self.datareadtime['currenttime']) #time that time field was read
    localweeksecs = self._weeksecs(localtimearray)
    remoteweeksecs = self._weeksecs(self.data['currenttime'])
    directdifference = abs(localweeksecs - remoteweeksecs)
    wrappeddifference = abs(self.DAYSECS * 7 - directdifference) #compute the difference on rollover
    self.timeerr = min(directdifference, wrappeddifference)
    logging.debug("Local time %i, remote time %i, error %i"%(localweeksecs,remoteweeksecs,self.timeerr))

    if self.timeerr > self.DAYSECS:
        raise hmControllerTimeError("C%2d Incorrect day : local is %s, sensor is %s" % (self._address, localtimearray[CURRENT_TIME_DAY], self.data['currenttime'][CURRENT_TIME_DAY]))

    if (self.timeerr > TIME_ERR_LIMIT):
        raise hmControllerTimeError("C%2d Time Error %d greater than %d: local is %s, sensor is %s" % (self._address, self.timeerr, TIME_ERR_LIMIT, localweeksecs, remoteweeksecs))

  def _localtimearray(self, timenow = time.time()):
    #creates an array in heatmiser format for local time. Day 1-7, 1=Monday
    #input time.time() (not local)
    localtimenow = time.localtime(timenow)
    nowday = localtimenow.tm_wday + 1  #python tm_wday, range [0, 6], Monday is 0
    nowsecs = min(localtimenow.tm_sec, 59) #python tm_sec range[0, 61]
    
    return [nowday, localtimenow.tm_hour, localtimenow.tm_min, nowsecs]
  
  DAYSECS = 86400
  HOURSECS = 3600
  MINSECS = 60
  def _weeksecs(self, localtimearray):
    #calculates the time from the start of the week in seconds from a heatmiser time array
    return ( localtimearray[CURRENT_TIME_DAY] - 1 ) * self.DAYSECS + localtimearray[CURRENT_TIME_HOUR] * self.HOURSECS + localtimearray[CURRENT_TIME_MIN] * self.MINSECS + localtimearray[CURRENT_TIME_SEC]
  
#### External functions for printing data
  def display_heating_schedule(self):
    self.heat_schedule.display()
      
  def display_water_schedule(self):
    if not self.water_schedule is None:
      self.water_schedule.display()

  def printTarget(self):
      
    current_state = self.getTempState()
    
    if current_state == self.TEMP_STATE_OFF:
      return "controller off without frost protection"
    elif current_state == self.TEMP_STATE_OFF_FROST:
      return "controller off"
    elif current_state == self.TEMP_STATE_HOLIDAY:
      return "controller on holiday for %i hours" % self.holidayhours
    elif current_state == self.TEMP_STATE_FROST:
      return "controller in frost mode"
    elif current_state == self.TEMP_STATE_HELD:
      return "temp held for %i mins at %i"%(self.tempholdmins, self.setroomtemp)
    elif current_state == self.TEMP_STATE_OVERRIDDEN:
      locatimenow = self._localtimearray()
      nexttarget = self.heat_schedule.getNextScheduleItem(locatimenow)
      return "temp overridden to %0.1f until %02d:%02d" % (self.setroomtemp, nexttarget[1], nexttarget[2])
    elif current_state == self.TEMP_STATE_PROGRAM:
      locatimenow = self._localtimearray()
      nexttarget = self.heat_schedule.getNextScheduleItem(locatimenow)
      return "temp set to %0.1f until %02d:%02d" % (self.setroomtemp, nexttarget[1], nexttarget[2])
    
  def _check_data_age(self, maxage, *fieldnames):
    #field data age is not more than maxage (in seconds)
    #return False if old, True if recent
    if len(fieldnames) == 0:
      raise ValueError("Must list at least one field")
    
    for fieldname in fieldnames:
      if not self._check_data_present(fieldname):
        return False
      if time.time() - self.datareadtime[fieldname] > maxage:
        logging.warning("C%i data item %s too old"%(self._address, fieldname))
        return False
    return True
    
  def _check_data_present(self, *fieldnames):
    if len(fieldnames) == 0:
      raise ValueError("Must list at least one field")

    for fieldname in fieldnames:
      if self.datareadtime[fieldname] == None:
        logging.warning("C%i data item %s not avaliable"%(self._address, fieldname))
        return False
    return True
        
#### External functions for getting data

  def isHotWater(self):
    #returns True if stat is a model with hotwater control, False otherwise
    return self._expected_model == 'prt_hw_model'

  TEMP_STATE_OFF = 0  #thermostat display is off and frost protection disabled
  TEMP_STATE_OFF_FROST = 1 #thermostat display is off and frost protection enabled
  TEMP_STATE_FROST = 2 #frost protection enabled indefinitely
  TEMP_STATE_HOLIDAY = 3 #holiday mode, frost protection for a period
  TEMP_STATE_HELD = 4 #temperature held for a number of hours
  TEMP_STATE_OVERRIDDEN = 5 #temperature overridden until next program time
  TEMP_STATE_PROGRAM = 6 #following program
  
  def getTempState(self):
    if not self._check_data_present('onoff','frostprot','holidayhours','runmode','tempholdmins','setroomtemp'):
      if self._autoreadall is True:
        self.hmReadAll()
      else:
        raise ValueError("Need to read all before getting temp state")
        
    if not self._check_data_age(self._max_age_variables, 'onoff','holidayhours','runmode','tempholdmins','setroomtemp'):
      if self._autoreadall is True:
        self.hmReadVariables()
      else:
        raise ValueError("Vars to old to get temp state")
    
    if self.onoff == WRITE_ONOFF_OFF and self.frostprot == READ_FROST_PROT_OFF:
      return self.TEMP_STATE_OFF
    elif self.onoff == WRITE_ONOFF_OFF and self.frostprot == READ_FROST_PROT_ON:
      return self.TEMP_STATE_OFF_FROST
    elif self.holidayhours != 0:
      return self.TEMP_STATE_HOLIDAY
    elif self.runmode == WRITE_RUNMODE_FROST:
      return self.TEMP_STATE_FROST
    elif self.tempholdmins != 0:
      return self.TEMP_STATE_HELD
    else:
    
      if not self._check_data_age(self._max_age_time, 'currenttime'):
        currenttime = self.readTime()
      
      locatimenow = self._localtimearray()
      scheduletarget = self.heat_schedule.getCurrentScheduleItem(locatimenow)

      if scheduletarget[SCH_ENT_TEMP] != self.setroomtemp:
        return self.TEMP_STATE_OVERRIDDEN
      else:
        return self.TEMP_STATE_PROGRAM

  ### UNTESTED OR EVEN CHECKED
  def getWaterState(self):
    #does runmode affect hot water state?
    if not self._check_data_present('onoff','holidayhours','hotwaterdemand'):
      if self._autoreadall is True:
        self.hmReadAll()
      else:
        raise ValueError("Need to read all before getting temp state")
        
    if not self._check_data_age(self._max_age_variables, 'onoff','holidayhours','hotwaterdemand'):
      if self._autoreadall is True:
        self.hmReadVariables()
      else:
        raise ValueError("Vars to old to get temp state")
    
    if self.onoff == WRITE_ONOFF_OFF:
      return self.TEMP_STATE_OFF
    elif self.holidayhours != 0:
      return self.TEMP_STATE_HOLIDAY
    else:
    
      if not self._check_data_age(self._max_age_time, 'currenttime'):
        currenttime = self.readTime()
      
      locatimenow = self._localtimearray()
      scheduletarget = self.water_schedule.getCurrentScheduleItem(locatimenow)

      if scheduletarget[SCH_ENT_TEMP] != self.hotwaterdemand:
        return self.TEMP_STATE_OVERRIDDEN
      else:
        return self.TEMP_STATE_PROGRAM
        
  def getAirSensorType(self):
    if not self._check_data_present('sensorsavaliable'):
      return False

    if self.sensorsavaliable == READ_SENSORS_AVALIABLE_INT_ONLY or self.sensorsavaliable == READ_SENSORS_AVALIABLE_INT_FLOOR:
      return 1
    elif self.sensorsavaliable == READ_SENSORS_AVALIABLE_EXT_ONLY or self.sensorsavaliable == READ_SENSORS_AVALIABLE_EXT_FLOOR:
      return 2
    else:
      return 0
      
  def getAirTemp(self):
    #if not read before read sensorsavaliable field
    self.readField('sensorsavaliable',None) 
    
    if self.sensorsavaliable == READ_SENSORS_AVALIABLE_INT_ONLY or self.sensorsavaliable == READ_SENSORS_AVALIABLE_INT_FLOOR:
      return self.readField('airtemp', self._max_age_temp)
    elif self.sensorsavaliable == READ_SENSORS_AVALIABLE_EXT_ONLY or self.sensorsavaliable == READ_SENSORS_AVALIABLE_EXT_FLOOR:
      return self.readField('remoteairtemp', self._max_age_temp)
    else:
      raise ValueError("sensorsavaliable field invalid")
     
#### External functions for setting data

  def setHeatingSchedule(self, day, schedule):
    padschedule = self.heat_schedule.pad_schedule(schedule)
    self._adaptor.setField(self._address,self._protocol,day,padschedule)
    
  def setWaterSchedule(self, day, schedule):
    padschedule = self.water_schedule.pad_schedule(schedule)
    if day == 'all':
      self.setField('mon_water',padschedule)
      self.setField('tues_water',padschedule)
      self.setField('wed_water',padschedule)
      self.setField('thurs_water',padschedule)
      self.setField('fri_water',padschedule)
      self.setField('sat_water',padschedule)
      self.setField('sun_water',padschedule)
    else:
      self.setField(day,padschedule)

  def setTime(self) :
      """set time on controller to match current localtime on server"""
      timenow = time.time() + 0.5 #allow a little time for any delay in setting
      return self.setField('currenttime',self._localtimearray(timenow))
      
#general field setting

  def setField(self,fieldname,payload):
    #set a field (single member of fields) to a state or payload. Defined for all field lengths.
    fieldinfo = fields[self._fieldnametonum[fieldname]]
    
    if len(fieldinfo) < FIELD_WRITE + 1 or fieldinfo[FIELD_WRITE] != 'W':
        #check that write is part of field info and is 'W'
        raise ValueError("setField: field isn't writeable")        
               
    self._checkPayloadValues(payload, fieldinfo)

    if fieldinfo[FIELD_LEN] == 1:
        payload = [payload]
    elif fieldinfo[FIELD_LEN] == 2:
        pay_lo = (payload & BYTEMASK)
        pay_hi = (payload >> 8) & BYTEMASK
        payload = [pay_lo, pay_hi]
    try:
        print payload
        self._adaptor.hmWriteToController(self._address, self._protocol, fieldinfo[FIELD_ADD], fieldinfo[FIELD_LEN], payload)
    except:
        logging.info("C%i failed to set field %s to %s"%(self._address, fieldname.ljust(FIELD_NAME_LENGTH), ', '.join(str(x) for x in payload)))
        raise
    else:
        logging.info("C%i set field %s to %s"%(self._address, fieldname.ljust(FIELD_NAME_LENGTH), ', '.join(str(x) for x in payload)))
    
    self.lastreadtime = time.time()
    
    ###should really be handled by a specific overriding function, rather than in here.
    #handle odd effect on WRITE_hotwaterdemand_PROG
    if fieldname == 'hotwaterdemand':
      if value == WRITE_HOTWATERDEMAND_PROG: #returned to program so outcome is unknown
        self.datareadtime[field] = None
        return None
      elif value == WRITE_HOTWATERDEMAND_OFF: #if overridden off store the off read value
        value = READ_HOTWATERDEMAND_OFF
    
    self._procpartpayload(payload,fieldname,fieldname)
    
  def _checkPayloadValues(self, payload, fieldinfo):
      #check the payload matches field details
      
      if fieldinfo[FIELD_LEN] in [1, 2] and not isinstance(payload, (int, long)):
          #one or two byte field, not single length payload
          raise TypeError("setField: invalid requested value")
      elif fieldinfo[FIELD_LEN] > 2 and len(payload) != fieldinfo[FIELD_LEN]:
          #greater than two byte field, payload length must match field length
          raise ValueError("setField: invalid payload length")
  
      #checks the payload matches the ranges if ranges are defined 
      ranges = fieldinfo[FIELD_RANGE]
      if ranges != []:
          if isinstance(payload, (int, long)):
              if ( payload < ranges[0] or payload > ranges[1] ):
                  raise ValueError("setField: payload out of range")
          else:
              for i, item in enumerate(payload):
                  range = ranges[i % len(ranges)]
                  if item < range[0] or item > range[1]:
                      raise ValueError("setField: payload out of range")
  
#overriding      
      
  def setTemp(self, temp) :
    #sets the temperature demand overriding the program. Believe it returns at next prog change.
  
    #check hold temp not applied
    if self.readField('tempholdmins') == 0:
      return self._adaptor.setField(self._address,self._protocol,'setroomtemp',temp)
    else:
      logging.warn("%i address, temp hold applied so won't set temp"%(self._address))

  def releaseTemp(self) :
    #release SetTemp back to the program, but only if temp isn't held
    if self.readField('tempholdmins') == 0:
      return self._adaptor.setField(self._address,self._protocol,'tempholdmins',0)
    else:
      logging.warn("%i address, temp hold applied so won't remove set temp"%(self._address))     

  def holdTemp(self, minutes, temp) :
    #sets the temperature demand overrding the program for a set time. Believe it then returns to program.
    self._adaptor.setField(self._address,self._protocol,'setroomtemp',temp)
    return self._adaptor.setField(self._address,self._protocol,'tempholdmins',minutes)
    #didn't stay on if did minutes followed by temp.
    
  def releaseHoldTemp(self) :
    #release SetTemp or HoldTemp back to the program
    return self._adaptor.setField(self._address,self._protocol,'tempholdmins',0)
    
  def setHoliday(self, hours) :
    #sets holiday up for a defined number of hours
    return self._adaptor.setField(self._address,self._protocol,'holidayhours',hours)
  
  def releaseHoliday(self) :
    #cancels holiday mode
    return self._adaptor.setField(self._address,self._protocol,'holidayhours',0)

#onoffs

  def setOn(self):
    return self._adaptor.setField(self._address,self._protocol,'onoff',WRITE_ONOFF_ON)
  def setOff(self):
    return self._adaptor.setField(self._address,self._protocol,'onoff',WRITE_ONOFF_OFF)
    
  def setHeat(self):
    return self._adaptor.setField(self._address,self._protocol,'runmode',WRITE_RUNMODE_HEATING)
  def setFrost(self):
    return self._adaptor.setField(self._address,self._protocol,'runmode',WRITE_RUNMODE_FROST)
    
  def setLock(self):
    return self._adaptor.setField(self._address,self._protocol,'keylock',WRITE_KEYLOCK_ON)
  def setUnlock(self):
    return self._adaptor.setField(self._address,self._protocol,'keylock',WRITE_KEYLOCK_OFF)
  
#other
#set floor limit
#set holiday

class hmBroadcastController(hmController):
  #create a controller that only broadcasts
  def __init__(self, network, long_name):
    settings = {'address':BROADCAST_ADDR,'display_order': 0, 'long_name': long_name,'protocol':DEFAULT_PROTOCOL,'expected_model':False,'expected_prog_mode':DEFAULT_PROG_MODE}
    super(hmBroadcastController, self).__init__(network, settings)
  
  ##add methods to block or remove get functions



# Catch all, not used directly in code
class hmError(RuntimeError):
    pass

# Raise this when response from a device is wrong
class hmResponseError(hmError):
    pass

# Specifically when CRC fails check. This is the most common response error.
class hmResponseErrorCRC(hmResponseError):
    pass

# Raise this when controller time is outside of boundarys
class hmControllerTimeError(hmError):
    pass
    
# Raise this when init fails.
class HeatmiserControllerSetupInitError(hmError):
    pass


#
# Ian Horsley 2018

#
# Sets time and gets information on all controllers
#

# Assume Python 2.7.x
#

# Protocols
HMV2_ID = 2
HMV3_ID = 3
DEFAULT_PROTOCOL = HMV3_ID

BYTEMASK = 0xff

DONT_CARE_LENGTH = 1
TIME_ERR_LIMIT = 10

#
# HM Version 3 Magic Numbers
#

# Master must be in range [0x81,0xa0] = [129,160]
MASTER_ADDR_MIN = 0x81
MASTER_ADDR_MAX = 0xa0
SLAVE_ADDR_MIN = 1
SLAVE_ADDR_MAX = 32

MAX_FRAME_RESP_LENGTH = 159 # NB max return is 75 in 5/2 mode or 159 in 7day mode
MIN_FRAME_RESP_LENGTH = 7
FRAME_WRITE_RESP_LENGTH = 7
MIN_FRAME_READ_RESP_LENGTH = 11
MIN_FRAME_SEND_LENGTH = 10
MAX_PAYLOAD_SEND_LENGTH = 100
CRC_LENGTH = 2

# Define magic numbers used in messages
FUNC_READ  = 0
FUNC_WRITE = 1

BROADCAST_ADDR = 0xff
RW_LENGTH_ALL = 0xffff
DCB_START = 0x00

SET_TEMP_ADDR = 18
KEY_LOCK_ADDR = 22
HOL_HOURS_LO_ADDR = 24
HOL_HOURS_HI_ADDR = 25
CUR_TIME_ADDR = 43

KEY_LOCK_UNLOCK = 0
KEY_LOCK_LOCK = 1

# Define frame send contents
FS_DEST_ADDR = 0
FS_LEN = 1
FS_SOURCE_ADDR = 2
FS_FUNC_CODE = 3

# Define frame response contents
FR_DEST_ADDR = 0
FR_LEN_LOW = 1
FR_LEN_HIGH = 2
FR_SOURCE_ADDR = 3
FR_FUNC_CODE = 4
FR_START_LOW = 5 # this field and following on present if function code read
FR_START_HIGH = 6
FR_CONT_LEN_LOW = 7
FR_CONT_LEN_HIGH = 8
FR_CONTENTS = 9

# Define payload response contents
PL_LEN_HIGH = 0
PL_LEN_LOW = 1 
PL_VENDOR_ID = 2
#PL_LIMITS = 3
PL_MODEL = 4
PL_PROG_MODE = 16

# Define DCB write addresses
UNIQUE_WRITEABLE_ADDRESSES = [17,18,19,21,22,23,24,32,42,43]
UNIQUE_WRITEABLE_ABOVE = 47

DCB_ON_OFF = 21
DCB_RUN_MODE = 23

UNIQUE_ADD_MODEL = 4

FIELD_NAME_LENGTH = 13
MAX_UNIQUE_ADDRESS = 298
#name, unique_address,length,divisor, valid range, writeable]
FIELD_NAME = 0
FIELD_ADD = 1
FIELD_LEN = 2
FIELD_DIV = 3
FIELD_RANGE = 4
FIELD_WRITE = 5
fields = [
['DCBlen',0,2,1,[]],
['vendor',2,1,1,[0,1]], #00 heatmiser, 01 OEM
['version',3,1,1,[]],
['model',4,1,1,[0,5]], # DT/DT-E/PRT/PRT-E 00/01/02/03
['tempformat',5,1,1,[0,1]], # 00 C, 01 F
['switchdiff',6,1,1,[1,3]],
['frostprot',7,1,1,[0,1]], #0=enable frost prot when display off, (opposite in protocol manual, but tested and user guide is correct)  (default should be enabled)
['caloffset',8,2,1,[]],
['outputdelay',10,1,1,[0,15]], # minutes (to prevent rapid switching)
['address',11,1,1,[SLAVE_ADDR_MIN,SLAVE_ADDR_MAX]],
['updwnkeylimit',12,1,1,[0,10]],  #limits use of up and down keys
['sensorsavaliable',13,1,1,[0,4]], #00 built in only, 01 remote air only, 02 floor only, 03 built in + floor, 04 remote + floor
['optimstart',14,1,1,[0,3]], # 0 to 3 hours, default 0
['rateofchange',15,1,1,[]], #number of minutes per degree to raise the temperature, default 20. Applies to the Wake and Return comfort levels (1st and 3rd)
['programmode',16,1,1,[0,1]], #0=5/2, 1= 7day
['frosttemp',17,1,1,[7,17],'W'], #default is 12, frost protection temperature
['setroomtemp',18,1,1,[5,35],'W'],
['floormaxlimit',19,1,1,[20,45],'W'],
['floormaxlimitenable',20,1,1,[0,1]], #1=enable
['onoff',21,1,1,[0,1],'W'], #1 = on
['keylock',22,1,1,[0,1],'W'], #1 = on
['runmode',23,1,1,[0,1],'W'],  #0 = heating mode, 1 = frost protection mode
['holidayhours',24,2,1,[0,720],'W'], #range guessed and tested, setting to 0 cancels hold and puts back to program 
['unknown',26,6,1,[]], # gap from 26 to 31
['tempholdmins',32,2,1,[0,5760],'W'], #range guessed and tested, setting to 0 cancels hold and puts setroomtemp back to program
['remoteairtemp',34,2,10,[]], #ffff if no sensor
['floortemp',36,2,10,[]], #ffff if no sensor
['airtemp',38,2,10,[]], #ffff if no sensor
['errorcode',40,1,1,[0,3]], # 0 is no error # errors, 0 built in, 1, floor, 2 remote
['heatingdemand',41,1,1,[0,1]], #0 none, 1 heating currently
['hotwaterdemand',42,1,1,[0,3],'W'], # read [0=off,1=on], write [0=as prog,1=override on,2=overide off]
['currenttime',43,4,1,[[1,7],[0,23],[0,59],[0,59]],'W'], #day (Mon - Sun), hour, min, sec.
#5/2 progamming #if hour = 24 entry not used
['wday_heat',47,12,1,[[0,24],[0,59],[5,35]],"W"], #hour, min, temp  (should minutes be only 0 and 30?)
['wend_heat',59,12,1,[[0,24],[0,59],[5,35]],"W"],
['wday_water',71,16,1,[[0,24],[0,59]],"W"], # pairs, on then off repeated, hour, min
['wend_water',87,16,1,[[0,24],[0,59]],"W"],
#7day progamming
['mon_heat',103,12,1,[[0,24],[0,59],[5,35]],"W"],
['tues_heat',115,12,1,[[0,24],[0,59],[5,35]],"W"],
['wed_heat',127,12,1,[[0,24],[0,59],[5,35]],"W"],
['thurs_heat',139,12,1,[[0,24],[0,59],[5,35]],"W"],
['fri_heat',151,12,1,[[0,24],[0,59],[5,35]],"W"],
['sat_heat',163,12,1,[[0,24],[0,59],[5,35]],"W"],
['sun_heat',175,12,1,[[0,24],[0,59],[5,35]],"W"],
['mon_water',187,16,1,[[0,24],[0,59]],"W"],
['tues_water',203,16,1,[[0,24],[0,59]],"W"],
['wed_water',219,16,1,[[0,24],[0,59]],"W"],
['thurs_water',235,16,1,[[0,24],[0,59]],"W"],
['fri_water',251,16,1,[[0,24],[0,59]],"W"],
['sat_water',267,16,1,[[0,24],[0,59]],"W"],
['sun_water',283,16,1,[[0,24],[0,59]],"W"]
]

CURRENT_TIME_DAY = 0
CURRENT_TIME_HOUR = 1
CURRENT_TIME_MIN = 2
CURRENT_TIME_SEC = 3

READ_FROST_PROT_ON = 0
READ_FROST_PROT_OFF = 1

READ_SENSORS_AVALIABLE_INT_ONLY = 00
READ_SENSORS_AVALIABLE_EXT_ONLY = 01
READ_SENSORS_AVALIABLE_FLOOR_ONLY = 02
READ_SENSORS_AVALIABLE_INT_FLOOR = 03
READ_SENSORS_AVALIABLE_EXT_FLOOR = 04

READ_PROGRAM_MODE_5_2 = 0
READ_PROGRAM_MODE_7 = 1

WRITE_ONOFF_OFF = 0
WRITE_ONOFF_ON = 1
WRITE_KEYLOCK_OFF = 0
WRITE_KEYLOCK_ON = 1

WRITE_RUNMODE_HEATING = 0
WRITE_RUNMODE_FROST = 1

WRITE_HOTWATERDEMAND_PROG = 0
WRITE_HOTWATERDEMAND_OVER_ON = 1
WRITE_HOTWATERDEMAND_OVER_OFF = 2
READ_HOTWATERDEMAND_OFF = 0
READ_HOTWATERDEMAND_ON = 1

#these maps are used to translate the unique address which is the same for all controllers to the DCB address for a specific controller
PROG_MODE_WEEK = 'week'
PROG_MODE_DAY = 'day'
DEFAULT_PROG_MODE = PROG_MODE_DAY #allows broadcast to program both modes.
PROG_MODES = {'week':0,'day':1}

#models
DEVICE_MODELS = {'prt_e_model': 3, 'prt_hw_model': 4, False: 0}
#PRT-E DCB map
DCB_INVALID = None
PRTEmap = {}
PRTEmap[PROG_MODE_WEEK] = list(reversed([(25,0),(31,DCB_INVALID),(41,6),(42,DCB_INVALID),(70,7),(MAX_UNIQUE_ADDRESS,DCB_INVALID)]))
PRTEmap[PROG_MODE_DAY] = list(reversed([(25,0),(31,DCB_INVALID),(41,6),(42,DCB_INVALID),(70,7),(102,DCB_INVALID),(186,39),(MAX_UNIQUE_ADDRESS,DCB_INVALID)]))
PRTHWmap = {}
PRTHWmap[PROG_MODE_WEEK] = list(reversed([(25,0),(31,DCB_INVALID),(102,6),(MAX_UNIQUE_ADDRESS,DCB_INVALID)]))
PRTHWmap[PROG_MODE_DAY] = list(reversed([(25,0),(31,DCB_INVALID),(MAX_UNIQUE_ADDRESS,6)]))
STRAIGHTmap = list([(MAX_UNIQUE_ADDRESS,0)])

#Fields for each stat type and mode, list of ranges.
#PRT-E fields
FIELDRANGES = {}
FIELDRANGES['prt_e_model'] = {}
FIELDRANGES['prt_e_model'][PROG_MODE_WEEK] = [['DCBlen','holidayhours'],['tempholdmins','heatingdemand'],['currenttime','wend_heat']]
FIELDRANGES['prt_e_model'][PROG_MODE_DAY] = [['DCBlen','holidayhours'],['tempholdmins','heatingdemand'],['currenttime','wend_heat'],['mon_heat','sun_heat']]
#PRT-HW fields
FIELDRANGES['prt_hw_model'] = {}
FIELDRANGES['prt_hw_model'][PROG_MODE_WEEK] = [['DCBlen','holidayhours'],['tempholdmins','wend_water']]
FIELDRANGES['prt_hw_model'][PROG_MODE_DAY] = [['DCBlen','holidayhours'],['tempholdmins','sun_water']]
#Full map
FIELDRANGES[False] = {PROG_MODE_WEEK: [['DCBlen','sun_water']],PROG_MODE_DAY: [['DCBlen','sun_water']]}




import logging

from hm_constants import *
from .exceptions import hmResponseError, hmResponseErrorCRC

### low level framing functions
    
def _hmFormReadFrame(destination, protocol, source, start, length) :
  return _hmFormFrame(destination, protocol, source, FUNC_READ, start, length, [])

# TODO check master address is in legal range
def _hmFormFrame(destination, protocol, source, function, start, length, payload) :
  """Forms a message payload, including CRC"""
  if protocol != HMV3_ID:
    raise ValueError("Protocol unknown")
  start_low = (start & BYTEMASK)
  start_high = (start >> 8) & BYTEMASK
  length_low = (length & BYTEMASK)
  length_high = (length >> 8) & BYTEMASK
  payload_length = len(payload)
  frame_length = MIN_FRAME_SEND_LENGTH
  if function == FUNC_WRITE:
    if length != payload_length:
      raise ValueError("Payload doesn't match length %s" % length)
    if length > MAX_PAYLOAD_SEND_LENGTH:
      raise ValueError("Payload to long %s" % length)
    frame_length = MIN_FRAME_SEND_LENGTH + payload_length
  msg = [destination, frame_length, source, function, start_low, start_high, length_low, length_high]
  if function == FUNC_WRITE:
    msg = msg + payload

  crc = crc16()
  msg = msg + crc.run(msg)
  return msg

def _hmCheckFrameCRC(protocol, data):
  """Takes frame with CRC and checks it is valid"""
  datalength = len(data)
  
  if protocol != HMV3_ID:
    raise ValueError("Protocol unknown")
  if datalength < 2 :
    raise hmResponseError("No CRC")

  checksum = data[len(data)-2:]
  rxmsg = data[:len(data)-2]
  
  crc = crc16() # Initialises the CRC
  expectedchecksum = crc.run(rxmsg)
  if expectedchecksum != checksum:
    raise hmResponseErrorCRC("CRC is incorrect")      

def _hmCheckResponseFrameLength(protocol, data, expectedLength):
  """Takes frame and checks length, must be a receive frame"""
  if protocol != HMV3_ID:
    raise ValueError("Protocol unknown")

  if (len(data) < MIN_FRAME_RESP_LENGTH):
    raise hmResponseError("Response length too short: %s %s"% (len(data), MIN_FRAME_RESP_LENGTH))

  frame_len_l = data[FR_LEN_LOW]
  frame_len_h = data[FR_LEN_HIGH]
  frame_len = (frame_len_h << 8) | frame_len_l
  func_code = data[FR_FUNC_CODE]
  
  if (len(data) != frame_len):
    raise hmResponseError("Frame length does not match header: %s %s" % (len(data), frame_len))

  if (expectedLength != RW_LENGTH_ALL and func_code == FUNC_READ and frame_len != MIN_FRAME_READ_RESP_LENGTH + expectedLength ):
    # Read response length is wrong
    raise hmResponseError("Response length %s not EXPECTED value %s + %s given read request" % (frame_len, MIN_FRAME_READ_RESP_LENGTH, expectedLength ))
  if (func_code == FUNC_WRITE and frame_len != FRAME_WRITE_RESP_LENGTH):
    # Reply to Write is always 7 long
    raise hmResponseError("Response length %s not EXPECTED value %s given write request" % (frame_len, FRAME_WRITE_RESP_LENGTH ))
      
def _hmCheckResponseFrameAddresses(protocol, source, destination, data):

  if protocol != HMV3_ID:
    raise ValueError("Protocol unknown")
    
  dest_addr = data[FR_DEST_ADDR]        
  source_addr = data[FR_SOURCE_ADDR]

  if (dest_addr < MASTER_ADDR_MIN or dest_addr > MASTER_ADDR_MAX):
    raise hmResponseError("Destination address out of valid range %i" % dest_addr)
  if (dest_addr != destination):
    raise hmResponseError("Destination address incorrect %i" % dest_addr)
  if (source_addr < SLAVE_ADDR_MIN or source_addr > SLAVE_ADDR_MAX):
    raise hmResponseError("Source address out of valid range %i" % source_addr)
  if (source_addr != source):
    raise hmResponseError("Source address does not match %i" % source_addr)
      
def _hmCheckResponseFrameFunc(protocol, expectedFunction, data):

  if protocol != HMV3_ID:
    raise ValueError("Protocol unknown")

  func_code = data[FR_FUNC_CODE]

  if (func_code != FUNC_WRITE and func_code != FUNC_READ):
    raise hmResponseError("Unknown function  code: %i" % (func_code))
  if (func_code != expectedFunction):
    raise hmResponseError("Function  code was not as expected: %i" % (func_code))
    
def _hmVerifyWriteAck(protocol, source, destination, data) :
  """Verifies message response to write is correct"""
  return _hmVerifyResponse(protocol, source, destination, FUNC_WRITE, DONT_CARE_LENGTH, data)
 
def _hmVerifyResponse(protocol, source, destination, expectedFunction, expectedLength, data) :
  """Verifies response frame appears legal"""
  try:
    # check CRC
    _hmCheckFrameCRC(protocol, data)
    # check length
    _hmCheckResponseFrameLength(protocol, data, expectedLength)
    # check addresses
    _hmCheckResponseFrameAddresses(protocol, source, destination, data)
    # check function
    _hmCheckResponseFrameFunc(protocol, expectedFunction, data)
  except hmResponseError as e:
    logging.warning("C%s Invalid Response: %s: %s" % (source, str(e), data))
    raise
    
  ## missing check that it is valid for this type of controller. Use DCBUnique function not false.
  ## although if needed should be in devices and not in framing

# Believe this is known as CCITT (0xFFFF)
# This is the CRC function converted directly from the Heatmiser C code
# provided in their API
class crc16:
    LookupHigh = [
    0x00, 0x10, 0x20, 0x30, 0x40, 0x50, 0x60, 0x70,
    0x81, 0x91, 0xa1, 0xb1, 0xc1, 0xd1, 0xe1, 0xf1
    ]
    LookupLow = [
    0x00, 0x21, 0x42, 0x63, 0x84, 0xa5, 0xc6, 0xe7,
    0x08, 0x29, 0x4a, 0x6b, 0x8c, 0xad, 0xce, 0xef
    ]
    def __init__(self):
        self.high = BYTEMASK
        self.low = BYTEMASK

    def Update4Bits(self, val):
        # Step one, extract the Most significant 4 bits of the CRC register
        #print "val is %d" % (val)
        t = self.high>>4
        #print "t is %d" % (t)

        # XOR in the Message Data into the extracted bits
        t = t^val
        #print "t is %d" % (t)

        # Shift the CRC Register left 4 bits
        self.high = (self.high << 4)|(self.low>>4)
        self.high = self.high & BYTEMASK    # force char
        self.low = self.low <<4
        self.low = self.low & BYTEMASK  # force char

        # Do the table lookups and XOR the result into the CRC tables
        #print "t for lookup is %d" % (t)
        self.high = self.high ^ self.LookupHigh[t]
        self.high = self.high & BYTEMASK    # force char
        self.low  = self.low  ^ self.LookupLow[t]
        self.low = self.low & BYTEMASK  # force char
        #print "high is %d Low is %d" % (self.high, self.low)

    def CRC16_Update(self, val):
        self.Update4Bits(val>>4) # High nibble first
        self.Update4Bits(val & 0x0f) # Low nibble

    def run(self, message):
        """Calculates a CRC"""
        for c in message:
            #print c
            self.CRC16_Update(c)
        #print "CRC is Low %d High  %d" % (self.low, self.high)
        return [self.low, self.high]



#!/usr/bin/env python
#
# Ian Horsley 2018

#
# Sets a bunch of different configurations on stats
#
import logging

from heatmisercontroller.logging_setup import initialize_logger
from heatmisercontroller.hm_constants import *
from heatmisercontroller.network import *

initialize_logger('logs', logging.INFO, True)

hmn1 = HeatmiserNetwork()

#hmn1.hmSetTemp("Kit",25)
#hmn1.Kit.holdTemp(30,21) #mins, temp
#hmn1.Kit.releaseHoldTemp()

#hmn1.hmSetTemp("Cons",25)
#hmn1.hmReleaseTemp("Cons")
#hmn1.All.setField('holidayhours',96)
hmn1.All.releaseHoliday()

hmn1.Kit.setField('hotwaterdemand',WRITE_HOTWATERDEMAND_PROG)

#hmn1.hmSetField(BROADCAST_ADDR,'runmode',WRITE_RUNMODE_HEATING)
hmn1.All.setOn()
#hmn1.hmUpdateTime(2)
#hmn1.controllerByName('B1').setOff()
#hmn1.hmSetField(BROADCAST_ADDR,HMV3_ID,'onoff',WRITE_ONOFF_ON)  
#hmn1.hmSetField('Cons',HMV3_ID,'onoff',WRITE_ONOFF_OFF)

hmn1.All.setField('runmode',WRITE_RUNMODE_HEATING)

#hmn1.B2.setTemp(24)
#hmn1.Kit.setTemp(24)
#hmn1.B1.setTemp(24)
#hmn1.B2.setTemp(24)
#hmn1.Cons.setTemp(24)

#hmn1.All.setTemp(24)

hmn1.Kit.releaseTemp()
hmn1.B1.releaseTemp()
hmn1.B2.releaseTemp()
hmn1.Cons.releaseTemp()

downInactive = 20
downActive = 19
frost = 12
dayfrost = [7,0,frost]

wkday_zearly =    [5,00,downActive,8,30,frost,15,00,downInactive,21,00,frost]
wkday_zlate =     [7,00,downInactive,12,00,frost,17,00,downInactive,21,30,frost]
wkday_ztraining = [7,00,downActive,8,30,frost,16,30,downInactive,21,00,frost]
wkend_zearly =    [5,00,downInactive,21,30,frost]
wkend_zlate =     [7,00,downInactive,21,30,frost]
wkend_zoff =      [7,00,downInactive,21,30,frost]
hmn1.Kit.setHeatingSchedule('mon_heat',wkend_zearly)
hmn1.Kit.setHeatingSchedule('tues_heat',wkend_zearly)
hmn1.Kit.setHeatingSchedule('wed_heat',wkday_ztraining)
hmn1.Kit.setHeatingSchedule('thurs_heat',wkend_zearly)
hmn1.Kit.setHeatingSchedule('fri_heat',wkend_zearly)
hmn1.Kit.setHeatingSchedule('sat_heat',wkend_zoff)
hmn1.Kit.setHeatingSchedule('sun_heat',wkend_zoff)

eveningwater = [17,30,18,0]
nowater = []
hmn1.Kit.setWaterSchedule('all',eveningwater)

upSleep = 16
upAwake = 18
upwkday_zearly =  [5,0,upAwake,7,00,frost,20,30,upAwake,21,30,upSleep]
upwkday_zlate =   [7,0,upAwake,8,00,frost,21,30,upAwake,22,30,upSleep]
upwkday =         [7,0,upAwake,8,30,frost,21,00,upAwake,21,30,upSleep]
upwkend_zearly =  [5,0,upAwake,8,00,frost,20,30,upAwake,21,30,upSleep]
upwkend_zlate =   [7,0,upAwake,9,30,frost,21,30,upAwake,22,30,upSleep]
upwkend =         [7,0,upAwake,9,30,frost,21,00,upAwake,22,00,upSleep]
dayfrost = [7,0,frost]

hmn1.B1.setHeatingSchedule('mon_heat',dayfrost)
hmn1.B1.setHeatingSchedule('tues_heat',dayfrost)
hmn1.B1.setHeatingSchedule('wed_heat',dayfrost)
hmn1.B1.setHeatingSchedule('thurs_heat',dayfrost)
hmn1.B1.setHeatingSchedule('fri_heat',dayfrost)
hmn1.B1.setHeatingSchedule('sat_heat',dayfrost)
hmn1.B1.setHeatingSchedule('sun_heat',dayfrost)

hmn1.B2.setHeatingSchedule('mon_heat',dayfrost)
hmn1.B2.setHeatingSchedule('tues_heat',dayfrost)
hmn1.B2.setHeatingSchedule('wed_heat',dayfrost)
hmn1.B2.setHeatingSchedule('thurs_heat',dayfrost)
hmn1.B2.setHeatingSchedule('fri_heat',dayfrost)
hmn1.B2.setHeatingSchedule('sat_heat',dayfrost)
hmn1.B2.setHeatingSchedule('sun_heat',dayfrost)

#hmn1.hmSetFields('Kit','wday_heat',[7,0,19,9,30,10,17,0,19,21,30,10])
#hmn1.hmSetFields('B1',HMV3_ID,'wday_heat',[7,0,18,8,30,10,20,30,18,22,0,16])
#hmn1.hmSetFields('B2',HMV3_ID,'wday_heat',[7,0,19,8,30,10,20,30,19,22,0,16])
hmn1.controllerByName('Cons').setField('wday_heat',[9,0,12,21,30,10,24,0,5,24,0,5])

#hmn1.hmSetFields('Kit','wend_heat',[7,0,19,21,30,10,24,0,5,24,0,5])
#hmn1.hmSetFields('B1',HMV3_ID,'wend_heat',[7,0,18,9,30,10,20,30,18,22,0,16])
#hmn1.hmSetFields('B2',HMV3_ID,'wend_heat',[7,0,19,9,30,10,20,30,19,22,0,16])
hmn1.controllerByName('Cons').setField('wend_heat',[9,0,12,21,30,10,24,0,5,24,0,5])

hmn1.Sit.setHeatingSchedule('wday_heat',dayfrost)
hmn1.Sit.setHeatingSchedule('wend_heat',dayfrost)

#hmn1.hmSetFields("Kit",'wday_water',[7,0,8,0,16,0,17,0,24,0,24,0,24,0,24,0])
#hmn1.hmSetFields("Kit",'wday_water',[24,0,24,0,24,0,24,0,24,0,24,0,24,0,24,0])
#hmn1.hmSetFields("Kit",'wend_water',[8,0,9,0,18,0,19,0,24,0,24,0,24,0,24,0])

#hmn1.hmSetField("Kit",'frosttemp',9)
#hmn1.hmSetField("B1",'frosttemp',9)
#hmn1.hmSetField("B2",'frosttemp',9)
#hmn1.hmSetField("Cons",'frosttemp',9)

